@@ -1,31 +1,51 @@
 /* common.c - Functions that are common to server and clinet
  * Tim Martin
  */
-/***********************************************************
-        Copyright 1998 by Carnegie Mellon University
 
-                      All Rights Reserved
-
-Permission to use, copy, modify, and distribute this software and its
-documentation for any purpose and without fee is hereby granted,
-provided that the above copyright notice appear in all copies and that
-both that copyright notice and this permission notice appear in
-supporting documentation, and that the name of CMU not be
-used in advertising or publicity pertaining to distribution of the
-software without specific, written prior permission.
-
-CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
-ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
-CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
-ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
-WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
-ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
-SOFTWARE.
-******************************************************************/
+/* 
+ * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer. 
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in
+ *    the documentation and/or other materials provided with the
+ *    distribution.
+ *
+ * 3. The name "Carnegie Mellon University" must not be used to
+ *    endorse or promote products derived from this software without
+ *    prior written permission. For permission or any other legal
+ *    details, please contact  
+ *      Office of Technology Transfer
+ *      Carnegie Mellon University
+ *      5000 Forbes Avenue
+ *      Pittsburgh, PA  15213-3890
+ *      (412) 268-4387, fax: (412) 268-7395
+ *      tech-transfer@andrew.cmu.edu
+ *
+ * 4. Redistributions of any form whatsoever must retain the following
+ *    acknowledgment:
+ *    "This product includes software developed by Computing Services
+ *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
+ *
+ * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
+ * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
+ * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
+ * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
+ * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
+ * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ */
 
 #include <config.h>
+#include <stdio.h>
 #include <limits.h>
-#ifdef HAVE_VSYSLOG
+#ifdef HAVE_SYSLOG
 #include <syslog.h>
 #endif
 #include <stdarg.h>
@@ -34,8 +54,6 @@ SOFTWARE.
 #include <saslplug.h>
 #include "saslint.h"
 #ifdef WIN32
-/* for snprintf */
-#include <stdio.h>
 /* need to handle the fact that errno has been defined as a function
    in a dll, not an extern int */
 # ifdef errno
@@ -48,6 +66,8 @@ SOFTWARE.
 
 static const char build_ident[] = "$Build: libsasl " PACKAGE "-" VERSION " $";
 
+void *dispose_mutex = NULL;
+
 void (*_sasl_client_cleanup_hook)(void) = NULL;
 void (*_sasl_server_cleanup_hook)(void) = NULL;
 int (*_sasl_client_idle_hook)(sasl_conn_t *conn) = NULL;
@@ -114,42 +134,50 @@ int _sasl_strdup(const char *in, char **out, int *outlen)
 int sasl_encode(sasl_conn_t *conn, const char *input, unsigned inputlen,
 		char **output, unsigned *outputlen)
 {
-  if (! conn || ! input || ! output || ! outputlen)
-    return SASL_FAIL;
-  if (conn->oparams.encode==NULL)
-  {
-    /* just copy the string, no encryption */
-    *output = sasl_ALLOC(inputlen+1);
-    if (! *output) return SASL_NOMEM;
-    memcpy(*output, input, inputlen);
-    *outputlen = inputlen;
-    (*output)[inputlen] = '\0'; /* sanity for stupid clients */
-    return SASL_OK;
-  } else {
-    return conn->oparams.encode(conn->context, input,
-				inputlen, output, outputlen);
-  }
+    int result;
+    if (! conn || ! input || ! output || ! outputlen)
+	return SASL_FAIL;
+
+  
+
+    if (conn->oparams.encode==NULL)
+    {
+	/* just copy the string, no encryption */
+	*output = sasl_ALLOC(inputlen+1);
+	if (! *output) return SASL_NOMEM;
+	memcpy(*output, input, inputlen);
+	*outputlen = inputlen;
+	(*output)[inputlen] = '\0'; /* sanity for stupid clients */
+	return SASL_OK;
+    } else {
+
+	result = conn->oparams.encode(conn->context, input,
+				      inputlen, output, outputlen);
+	return result;
+    }
 }
 
 int sasl_decode(sasl_conn_t *conn, const char *input, unsigned inputlen,
 		char **output, unsigned *outputlen)
 {
-  if (! conn || ! input || ! output || ! outputlen)
-    return SASL_FAIL;
-  if (conn->oparams.decode==NULL)
-  {
-    /* just copy the string, no encryption */
-    *output = sasl_ALLOC(inputlen + 1);
-    if (! *output) return SASL_NOMEM;
-    memcpy(*output, input, inputlen);
-    *outputlen = inputlen;
-    (*output)[inputlen] = '\0'; /* sanity for stupid clients */
-    return SASL_OK;
-
-  } else {
-    return conn->oparams.decode(conn->context, input, inputlen,
-				output, outputlen);
-  }
+    int result;
+    if (! conn || ! input || ! output || ! outputlen)
+	return SASL_FAIL;
+    if (conn->oparams.decode==NULL)
+    {
+	/* just copy the string, no encryption */
+	*output = sasl_ALLOC(inputlen + 1);
+	if (! *output) return SASL_NOMEM;
+	memcpy(*output, input, inputlen);
+	*outputlen = inputlen;
+	(*output)[inputlen] = '\0'; /* sanity for stupid clients */
+	return SASL_OK;
+	
+    } else {
+	result = conn->oparams.decode(conn->context, input, inputlen,
+				      output, outputlen);
+	return result;
+    }
 }
 
 
@@ -172,6 +200,9 @@ void sasl_done(void)
 
   if (_sasl_client_cleanup_hook)
     _sasl_client_cleanup_hook();
+  
+  sasl_MUTEX_DISPOSE(dispose_mutex);
+  dispose_mutex = NULL;
 
   /* in case of another init/done */
   _sasl_server_cleanup_hook = NULL;
@@ -234,17 +265,40 @@ cleanup_mutex:
   return result;
 }
 
+int _sasl_common_init(void)
+{
+    if (!dispose_mutex)
+	dispose_mutex = sasl_MUTEX_NEW();
+    if (!dispose_mutex) return SASL_FAIL;
+
+    return SASL_OK;
+}
+
 /* dispose connection state, sets it to NULL
  *  checks for pointer to NULL
  */
 void sasl_dispose(sasl_conn_t **pconn)
 {
+  int result;
+
   if (! pconn) return;
   if (! *pconn) return;
 
+
+  /* serialize disposes. this is necessary because we can't
+     dispose of conn->mutex if someone else is locked on it
+     xxx there probably is a better way to do this */
+  result = sasl_MUTEX_LOCK(dispose_mutex);
+  if (result!=SASL_OK) return;
+
+  /* *pconn might have become NULL by now */
+  if (! (*pconn)) return;
+
   (*pconn)->destroy_conn(*pconn);
   sasl_FREE(*pconn);
   *pconn=NULL;
+
+  sasl_MUTEX_UNLOCK(dispose_mutex);
 }
 
 void _sasl_conn_dispose(sasl_conn_t *conn) {
@@ -360,7 +414,7 @@ int sasl_setprop(sasl_conn_t *conn, int propnum, const void *value)
 			      &str,
 			      NULL);
 	if (result != SASL_OK)
-	  return result;
+	    goto done_unlock;
       } else
 	str = NULL;
       if (conn->external.auth_id)
@@ -384,6 +438,8 @@ int sasl_setprop(sasl_conn_t *conn, int propnum, const void *value)
     default:
       result = SASL_BADPARAM;
   }
+
+ done_unlock:
   sasl_MUTEX_UNLOCK(conn->mutex);
   return result;
 }
@@ -515,8 +571,7 @@ _sasl_conn_getopt(void *context,
 			     len);
 }
 
-#ifdef HAVE_VSYSLOG
-
+#ifdef HAVE_SYSLOG
 /* this is the default logging */
 static int _sasl_syslog(void *context __attribute__((unused)),
 			int priority,
@@ -545,14 +600,7 @@ static int _sasl_syslog(void *context __attribute__((unused)),
     
     return SASL_OK;
 }
-
-
-#endif				/* HAVE_VSYSLOG */
-
-
-#ifdef WIN32
-#define MAX_VALUE_NAME	128
-#endif /* WIN32 */
+#endif				/* HAVE_SYSLOG */
 
 static int
 _sasl_getsimple(void *context,
@@ -562,33 +610,33 @@ _sasl_getsimple(void *context,
 {
   const char *userid;
   sasl_conn_t *conn;
-#ifdef WIN32
-  DWORD i;
-  BOOL rval;
-  static char sender[MAX_VALUE_NAME];
-#endif //WIN32
 
-  if (! context || ! result)
-    return SASL_BADPARAM;
+  if (! context || ! result) return SASL_BADPARAM;
 
   conn = (sasl_conn_t *)context;
 
   switch(id) {
   case SASL_CB_AUTHNAME:
-#ifndef WIN32
     userid = getenv("USER");
-#else
-    userid = getenv("USERNAME");
-#endif /* WIN32 */
     if (userid != NULL) {
-      *result = userid;
-      if (len) *len = strlen(userid);
-      return SASL_OK;
+	*result = userid;
+	if (len) *len = strlen(userid);
+	return SASL_OK;
+    }
+    userid = getenv("USERNAME");
+    if (userid != NULL) {
+	*result = userid;
+	if (len) *len = strlen(userid);
+	return SASL_OK;
     }
 #ifdef WIN32
     /* for win32, try using the GetUserName standard call */
-    else {
-	i=sizeof(sender);
+    {
+	DWORD i;
+	BOOL rval;
+	static char sender[128];
+	
+	i = sizeof(sender);
 	rval = GetUserName(sender, &i);
 	if ( rval) { /* got a userid */
 		*result = sender;
@@ -710,12 +758,12 @@ _sasl_getcallback(sasl_conn_t * conn,
 
   /* Otherwise, see if the library provides a default callback. */
   switch (callbackid) {
-#ifdef HAVE_VSYSLOG
+#ifdef HAVE_SYSLOG
   case SASL_CB_LOG:
     *pproc = (int (*)()) &_sasl_syslog;
     *pcontext = NULL;
     return SASL_OK;
-#endif /* HAVE_VSYSLOG */
+#endif /* HAVE_SYSLOG */
   case SASL_CB_GETPATH:
     *pproc = (int (*)()) &_sasl_getpath;
     *pcontext = NULL;
