@@ -1,28 +1,47 @@
 /* SASL server API implementation
  * Tim Martin
- * $Id: server.c,v 1.68 2000/02/24 01:36:22 leg Exp $
+ * $Id: server.c,v 1.77 2000/04/12 19:56:42 tmartin Exp $
  */
-/***********************************************************
-        Copyright 1998 by Carnegie Mellon University
 
-                      All Rights Reserved
-
-Permission to use, copy, modify, and distribute this software and its
-documentation for any purpose and without fee is hereby granted,
-provided that the above copyright notice appear in all copies and that
-both that copyright notice and this permission notice appear in
-supporting documentation, and that the name of CMU not be
-used in advertising or publicity pertaining to distribution of the
-software without specific, written prior permission.
-
-CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
-ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
-CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
-ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
-WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
-ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
-SOFTWARE.
-******************************************************************/
+/* 
+ * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ *
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer. 
+ *
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in
+ *    the documentation and/or other materials provided with the
+ *    distribution.
+ *
+ * 3. The name "Carnegie Mellon University" must not be used to
+ *    endorse or promote products derived from this software without
+ *    prior written permission. For permission or any other legal
+ *    details, please contact  
+ *      Office of Technology Transfer
+ *      Carnegie Mellon University
+ *      5000 Forbes Avenue
+ *      Pittsburgh, PA  15213-3890
+ *      (412) 268-4387, fax: (412) 268-7395
+ *      tech-transfer@andrew.cmu.edu
+ *
+ * 4. Redistributions of any form whatsoever must retain the following
+ *    acknowledgment:
+ *    "This product includes software developed by Computing Services
+ *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
+ *
+ * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
+ * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
+ * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
+ * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
+ * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
+ * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ */
 
 /* local functions/structs don't start with sasl
  */
@@ -68,9 +87,13 @@ extern int gethostname(char *, int);
  * sasl_setpass
  */
 
+/* if we've initialized the server sucessfully */
+static int _sasl_server_active = 0;
+
 static int _sasl_checkpass(sasl_conn_t *conn,
 			   const char *mech, const char *service, 
-			   const char *user, const char *pass);
+			   const char *user, const char *pass,
+			   const char **errstr);
 
 static int
 external_server_new(void *glob_context __attribute__((unused)),
@@ -230,22 +253,18 @@ typedef struct mech_list {
 } mech_list_t;
 
 typedef struct sasl_server_conn {
-  sasl_conn_t base; /* parts common to server + client */
-
-  char *user_realm; /* domain the user authenticating is in This is
-		      * usually simply their hostname */
-
-  int authenticated;
-  mechanism_t *mech; /* mechanism trying to use */
-  sasl_server_params_t *sparams;
-
+    sasl_conn_t base; /* parts common to server + client */
+    
+    char *user_realm; /* domain the user authenticating is in */
+    int authenticated;
+    mechanism_t *mech; /* mechanism trying to use */
+    sasl_server_params_t *sparams;
 } sasl_server_conn_t;
 
-static mech_list_t *mechlist; /* global var which holds the list */
+static mech_list_t *mechlist = NULL; /* global var which holds the list */
 
 static sasl_global_callbacks_t global_callbacks;
 
-
 /* set the password for a user
  *  conn        -- SASL connection
  *  user        -- user name
@@ -277,11 +296,11 @@ int sasl_setpass(sasl_conn_t *conn,
     sasl_getopt_t *getopt;
     void *context;
     const char *plainmech;
-
+     
     /* check params */
     if (!conn)
 	return SASL_BADPARAM;
-    
+     
     if (!mechlist) {
 	return SASL_FAIL;
     }
@@ -306,7 +325,7 @@ int sasl_setpass(sasl_conn_t *conn,
     tmpresult = _sasl_sasldb_set_pass(conn, user, pass, passlen, 
 				      s_conn->user_realm, flags, errstr);
     if (tmpresult != SASL_OK && tmpresult != SASL_NOCHANGE) {
-	result = SASL_FAIL;
+	result = tmpresult;
 	_sasl_log(conn, SASL_LOG_ERR, "PLAIN", tmpresult,
 #ifndef WIN32
 		  errno,
@@ -473,21 +492,29 @@ static int add_plugin(void *p, void *library) {
 static void server_done(void) {
   mechanism_t *m;
   mechanism_t *prevm;
-  m=mechlist->mech_list; /* m point to begging of the list */
 
-  while (m!=NULL)
+  if (mechlist != NULL)
   {
-    prevm=m;
-    m=m->next;
+      m=mechlist->mech_list; /* m point to begging of the list */
+
+      while (m!=NULL)
+      {
+	  prevm=m;
+	  m=m->next;
     
-    if (prevm->plug->glob_context!=NULL)
-      sasl_FREE(prevm->plug->glob_context);
-    if (prevm->library!=NULL)
-      _sasl_done_with_plugin(prevm->library);
-    sasl_FREE(prevm);    
+	  if (prevm->plug->glob_context!=NULL)
+	      sasl_FREE(prevm->plug->glob_context);
+	  if (prevm->library!=NULL)
+	      _sasl_done_with_plugin(prevm->library);
+	  sasl_FREE(prevm);    
+      }
+      _sasl_free_utils(&mechlist->utils);
+      sasl_FREE(mechlist);
+      mechlist = NULL;
   }
-  _sasl_free_utils(&mechlist->utils);
-  sasl_FREE(mechlist);
+
+  /* no longer active. fail on listmech's etc. */
+  _sasl_server_active = 0;
 }
 
 static int
@@ -524,7 +551,7 @@ static int load_config(const sasl_callback_t *verifyfile_cb)
      system */
   result = ((sasl_getpath_t *)(getpath_cb->proc))(getpath_cb->context,
 						  &path_to_config);
-  if (result!=SASL_OK) return result;
+  if (result!=SASL_OK) goto done;
   if (path_to_config == NULL) path_to_config = "";
 
   if ((c = strchr(path_to_config, ':'))) {
@@ -535,11 +562,17 @@ static int load_config(const sasl_callback_t *verifyfile_cb)
      for '\0' */
   len = strlen(path_to_config)+2+ strlen(global_callbacks.appname)+5+1;
 
-  if (len > PATH_MAX ) return SASL_FAIL;
+  if (len > PATH_MAX ) {
+      result = SASL_FAIL;
+      goto done;
+  }
 
   /* construct the filename for the config file */
   config_filename = sasl_ALLOC(len);
-  if (! config_filename) return SASL_NOMEM; 
+  if (! config_filename) {
+      result = SASL_NOMEM;
+      goto done;
+  }
 
   snprintf(config_filename, len, "%s/%s.conf", path_to_config, 
 	   global_callbacks.appname);
@@ -556,8 +589,9 @@ static int load_config(const sasl_callback_t *verifyfile_cb)
   if (result==SASL_OK)
     result=sasl_config_init(config_filename);
 
-  sasl_FREE(config_filename);
-  if (*path_to_config) { /* was path_to_config allocated? */
+ done:
+  if (config_filename) sasl_FREE(config_filename);
+  if ((path_to_config) && (*path_to_config)) { /* was path_to_config allocated? */
       sasl_FREE(path_to_config);
   }
 
@@ -606,6 +640,8 @@ int sasl_server_init(const sasl_callback_t *callbacks,
   _sasl_server_getsecret_hook = _sasl_db_getsecret;
   _sasl_server_putsecret_hook = _sasl_db_putsecret;
 
+  _sasl_server_cleanup_hook = &server_done;
+
   /* verify that the callbacks look ok */
   ret = verify_server_callbacks(callbacks);
   if (ret!=SASL_OK) return ret;
@@ -613,25 +649,28 @@ int sasl_server_init(const sasl_callback_t *callbacks,
   global_callbacks.callbacks = callbacks;
   global_callbacks.appname = appname;
 
+  /* allocate mechlist and set it to empty */
   mechlist=sasl_ALLOC(sizeof(mech_list_t));
   if (mechlist==NULL) return SASL_NOMEM;
+  mechlist->mech_list=NULL;
+  mechlist->mech_length=0;
+  ret=init_mechlist();
+  if (ret!=SASL_OK)
+    return ret;
 
   vf = _sasl_find_verifyfile_callback(callbacks);
 
   /* load config file if applicable */
   ret = load_config(vf);
-  if ((ret!=SASL_OK) && (ret!=SASL_CONTINUE)) return ret;
+  if ((ret!=SASL_OK) && (ret!=SASL_CONTINUE))
+  {
+      return ret;
+  }
 
   /* check db */
   ret = _sasl_server_check_db(vf);
 
   /* load plugins */
-  ret=init_mechlist();
-  if (ret!=SASL_OK)
-    return ret;
-  mechlist->mech_list=NULL;
-  mechlist->mech_length=0;
-
   add_plugin((void *)&external_server_init, NULL);
 
   ret=_sasl_get_mech_list("sasl_server_plug_init",
@@ -639,11 +678,13 @@ int sasl_server_init(const sasl_callback_t *callbacks,
 			  _sasl_find_verifyfile_callback(callbacks),
 			  &add_plugin);
 
+  if (ret == SASL_OK)
+      ret = _sasl_common_init();
   
   if (ret == SASL_OK)
   {
-      /* cleanup_hook shows if we're active or not. sasl_done() sets it back to null */
-      _sasl_server_cleanup_hook = &server_done;
+      /* _sasl_server_active shows if we're active or not. sasl_done() sets it back to 0 */
+      _sasl_server_active = 1;
       _sasl_server_idle_hook = &server_idle;
   }
 
@@ -742,7 +783,7 @@ int sasl_server_new(const char *service,
   serverconn->sparams=sasl_ALLOC(sizeof(sasl_server_params_t));
   if (serverconn->sparams==NULL) return SASL_NOMEM;
 
-  /* set util functions - need to do rest*/
+  /* set util functions - need to do rest */
   serverconn->sparams->utils=_sasl_alloc_utils(*pconn, &global_callbacks);
   if (serverconn->sparams->utils==NULL)
     return SASL_NOMEM;
@@ -759,11 +800,10 @@ int sasl_server_new(const char *service,
     result = _sasl_strdup(user_realm, &serverconn->user_realm, NULL);
   }
 
-  if (result!=SASL_OK)
-  {
-    _sasl_conn_dispose(*pconn);
-    sasl_FREE(*pconn);
-    *pconn = NULL;
+  if (result!=SASL_OK) {
+      _sasl_conn_dispose(*pconn);
+      sasl_FREE(*pconn);
+      *pconn = NULL;
   }
 
   return result;
@@ -873,12 +913,16 @@ int sasl_server_start(sasl_conn_t *conn,
     m=m->next;
   }
   
-  if (m==NULL)
-    return SASL_NOMECH;
+  if (m==NULL) {
+    result = SASL_NOMECH;
+    goto done;
+  }
 
   /* Make sure that we're willing to use this mech */
-  if (! mech_permitted(conn, m))
-    return SASL_NOMECH;
+  if (! mech_permitted(conn, m)) {
+    result = SASL_NOMECH;
+    goto done;
+  }
 
   s_conn->mech=m;
 
@@ -895,17 +939,18 @@ int sasl_server_start(sasl_conn_t *conn,
 					0,
 					&(conn->context),
 					errstr);
-  if (result != SASL_OK)
-    return result;
 
-  return s_conn->mech->plug->mech_step(conn->context,
-				       s_conn->sparams,
-				       clientin,
-				       clientinlen,
-				       serverout,
-				       (int *) serveroutlen,
-				       &conn->oparams,
-				       errstr);
+  if (result == SASL_OK)
+      result = s_conn->mech->plug->mech_step(conn->context,
+					     s_conn->sparams,
+					     clientin,
+					     clientinlen,
+					     serverout,
+					     (int *) serveroutlen,
+					     &conn->oparams,
+					     errstr);
+ done:
+  return result;
 }
 
 
@@ -970,7 +1015,8 @@ int sasl_server_step(sasl_conn_t *conn,
 		       s_conn->base.oparams.user, &canonuser, errstr);
 
 	if (ret == SASL_OK && canonuser != NULL) {
-	    sasl_FREE(s_conn->base.oparams.user);
+	    if (s_conn->base.oparams.user != NULL)
+		sasl_FREE(s_conn->base.oparams.user);
 	    s_conn->base.oparams.user = (char *) canonuser;
 	}
     }
@@ -1032,7 +1078,7 @@ int sasl_listmech(sasl_conn_t *conn,
   const char *mysep;
 
   /* if there hasn't been a sasl_sever_init() fail */
-  if (_sasl_server_cleanup_hook==NULL) return SASL_FAIL;
+  if (_sasl_server_active==0) return SASL_FAIL;
 
   if (! conn || ! result)
     return SASL_FAIL;
@@ -1111,42 +1157,23 @@ static int is_mech(const char *t, const char *m)
 /* returns OK if it's valid */
 static int _sasl_checkpass(sasl_conn_t *conn,
 			   const char *mech, const char *service,
-			   const char *user, const char *pass)
+			   const char *user, const char *pass,
+			   const char **errstr)
 {
     sasl_server_conn_t *s_conn = (sasl_server_conn_t *) conn;
     int result = SASL_NOMECH;
+    struct sasl_verify_password_s *v;
 
-#ifdef HAVE_PAM
-    if (is_mech(mech, "PAM")) {
-	result = _sasl_PAM_verify_password(conn, user, pass, service, NULL);
-    } else
-#endif
-#ifndef SASL_MINIMAL_SERVER
-    if (is_mech(mech, "passwd")) {
-	result = _sasl_passwd_verify_password(conn, user, pass, NULL);
-    } else
-    if (is_mech(mech, "shadow")) {
-	result = _sasl_shadow_verify_password(conn, user, pass, NULL);
-    } else
-#endif /* SASL_MINIMAL_SERVER */
-#ifdef HAVE_KRB
-    if (is_mech(mech, "kerberos_v4")) {
-	/* check against krb */
-	result = _sasl_kerberos_verify_password(conn, user, pass, 
-						service, NULL);
-    } else
-#endif
-#ifdef HAVE_PWCHECK
-    if (is_mech(mech, "pwcheck")) {
-	/* check against pwcheck daemon */
-	result = _sasl_pwcheck_verify_password(conn, user, pass, NULL);
-    } else
-#endif
-    if (is_mech(mech, "sasldb")) {
-	/* check sasl database */
-	result = _sasl_sasldb_verify_password(conn, user, pass, 
-					      s_conn->user_realm, NULL);
-    } else {
+    if (mech == NULL) mech = DEFAULT_PLAIN_MECHANISM;
+    for (v = _sasl_verify_password; v->name; v++) {
+	if (is_mech(mech, v->name)) {
+	    result = v->verify(conn, user, pass, 
+			       service, s_conn->user_realm, errstr);
+	    break;
+	}
+    }
+
+    if (result == SASL_NOMECH) {
 	/* no mechanism available ?!? */
 	_sasl_log(conn, SASL_LOG_ERR, NULL, 0, 0,
 		  "unrecognized plaintext verifier %s", mech);
@@ -1182,20 +1209,20 @@ int sasl_checkpass(sasl_conn_t *conn,
     sasl_getopt_t *getopt;
     void *context;
 
+    /* check params */
+    if (_sasl_server_active==0) return SASL_FAIL;
+    if ((conn == NULL) || (user == NULL) || (pass == NULL)) return SASL_BADPARAM;
+
     if (user == NULL) return SASL_NOUSER;
 
-    if (_sasl_getcallback(conn, SASL_CB_GETOPT, &getopt, &context) 
+    /* figure out how to check (i.e. PAM or /etc/passwd or kerberos or etc...) */
+    if (_sasl_getcallback(conn, SASL_CB_GETOPT, &getopt, &context)
 	    == SASL_OK) {
 	getopt(context, NULL, "pwcheck_method", &mech, NULL);
     }
 
-    if (mech == NULL) {
-	mech = DEFAULT_PLAIN_MECHANISM;
-    }
-
-    result = _sasl_checkpass(conn, mech, conn->service, user, pass);
-    if ( errstr !=NULL)
-      *errstr = NULL;
+    if (errstr != NULL) { *errstr = NULL; }
+    result = _sasl_checkpass(conn, mech, conn->service, user, pass, errstr);
 
     if (result == SASL_OK) {
 	result = _sasl_strdup(user, &(conn->oparams.authid), NULL);
